# Generated by Django 3.1.7 on 2021-06-29 12:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0092_auto_20210629_1124'),
    ]

    operations = [
        migrations.AlterField(
            model_name='task',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.project'),
        ),
    ]



# Generated by Django 4.2.16 on 2024-10-30 23:14

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0117_alter_tasknote_task_alter_ticketnote_ticket'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContractExclusionSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(blank=True, null=True)),
                ('is_active', models.BooleanField(default=True)),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.AddField(
            model_name='contract',
            name='contract_exclusion_set_id',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.CreateModel(
            name='ContractExclusionSetExcludedWorkType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contract_exclusion_set', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.contractexclusionset')),
                ('excluded_work_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.billingcode')),
            ],
        ),
        migrations.CreateModel(
            name='ContractExclusionSetExcludedRole',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('contract_exclusion_set', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.contractexclusionset')),
                ('excluded_role', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djautotask.role')),
            ],
        ),
        migrations.AddField(
            model_name='contractexclusionset',
            name='excluded_roles',
            field=models.ManyToManyField(related_name='excluded_role_sets', through='djautotask.ContractExclusionSetExcludedRole', to='djautotask.role'),
        ),
        migrations.AddField(
            model_name='contractexclusionset',
            name='excluded_work_types',
            field=models.ManyToManyField(related_name='excluded_work_type_sets', through='djautotask.ContractExclusionSetExcludedWorkType', to='djautotask.billingcode'),
        ),
        migrations.CreateModel(
            name='ContractExcludedRoleTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_contractexclusionsetrole',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.contractexclusionsetexcludedrole',),
        ),
        migrations.CreateModel(
            name='ContractExcludedWorkTypeTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_contractexclusionsetworktype',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.contractexclusionsetexcludedworktype',),
        ),
        migrations.CreateModel(
            name='ContractExclusionSetTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_contractexclusionset',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.contractexclusionset',),
        ),
    ]



# Generated by Django 4.2.16 on 2024-11-01 13:16

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0118_contractexclusionset_and_more'),
        ('djautotask', '0119_merge_20241024_1521'),
    ]

    operations = [
    ]



import logging
import os
import base64
from dateutil.parser import parse
from decimal import Decimal

from django.db import transaction, IntegrityError
from django.db.models import Q
from django.utils import timezone

from djautotask import api
from djautotask import models
from .api import ApiCondition as A, AutotaskRecordNotFoundError
from .utils import DjautotaskSettings

CREATED = 1
UPDATED = 2
SKIPPED = 3
FILE_UMASK = 0o022


logger = logging.getLogger(__name__)


class InvalidObjectException(Exception):
    """
    If for any reason an object can't be created (for example, it references
    an unknown foreign object, or is missing a required field), raise this
    so that the synchronizer can catch it and continue with other records.
    """
    pass


def log_sync_job(f):
    def wrapper(*args, **kwargs):
        sync_instance = args[0]
        created_count = updated_count = deleted_count = skipped_count = 0
        sync_job = models.SyncJob()
        sync_job.start_time = timezone.now()
        sync_job.entity_name = sync_instance.model_class.__bases__[0].__name__
        sync_job.synchronizer_class = sync_instance.__class__.__name__

        if sync_instance.full:
            sync_job.sync_type = 'full'
        else:
            sync_job.sync_type = 'partial'

        sync_job.save()

        try:
            created_count, updated_count, skipped_count, deleted_count = \
                f(*args, **kwargs)
            sync_job.success = True
        except Exception as e:
            error_msg = f'Failed to create object. The error was: {e}'
            sync_job.message = error_msg
            sync_job.success = False
            sync_job.save()
            raise
        finally:
            sync_job.end_time = timezone.now()
            sync_job.added = created_count
            sync_job.updated = updated_count
            sync_job.skipped = skipped_count
            sync_job.deleted = deleted_count
            sync_job.save()

        return created_count, updated_count, skipped_count, deleted_count

    return wrapper


class SyncResults:
    """Track results of a sync job."""

    def __init__(self):
        self.created_count = 0
        self.updated_count = 0
        self.skipped_count = 0
        self.deleted_count = 0
        self.synced_ids = set()


class ParentSynchronizer:

    def sync_related(self, instance):
        raise NotImplementedError

    def sync_children(self, *args):
        for synchronizer, query_params in args:
            created_count, updated_count, skipped_count, deleted_count = \
                synchronizer.callback_sync(
                    query_params
                )
            msg = '{} Child Sync - Created: {},' \
                  ' Updated: {}, Skipped: {}, Deleted: {}'.format(
                    synchronizer.model_class.__bases__[0].__name__,
                    created_count,
                    updated_count,
                    skipped_count,
                    deleted_count
                  )

            logger.info(msg)


class ChildSynchronizer:

    def _child_instance_ids(self, query_params):
        ids = self.model_class.objects.filter(
            ticket__id=query_params[1]
        ).values_list('id', flat=True)

        return set(ids)

    def _get_children(self, results, query_params):
        self._build_children_conditions(query_params)
        logger.info('Fetching {} records.'.format(
            self.model_class.__bases__[0].__name__))
        return self.fetch_records(results)

    def _build_children_conditions(self, query_params):
        parent_field, parent_id = query_params
        self.client.clear_conditions()
        self.client.add_condition(
            A(
                op='eq',
                field=parent_field,
                value=parent_id
            )
        )

    def callback_sync(self, query_params):
        results = SyncResults()

        # Set of IDs of all records prior to sync,
        # to find stale records for deletion.
        initial_ids = self._child_instance_ids(query_params)

        results = self._get_children(results, query_params)

        results.deleted_count = self.prune_stale_records(
            initial_ids,
            results.synced_ids
        )

        return results.created_count, results.updated_count, \
            results.skipped_count, results.deleted_count


class BatchQueryMixin:

    condition_pool = None
    condition_field_name = None
    batch_query_size = None
    client = None

    def __init__(self, full=False, *args, **kwargs):
        settings = DjautotaskSettings().get_settings()
        self.batch_query_size = settings.get('batch_query_size')
        super().__init__(full, *args, **kwargs)
        self._add_conditions()

    def _add_conditions(self):
        self.condition_pool = list(self.active_ids)
        self.client.add_condition(
            A(
                op='in',
                field=self.condition_field_name,
                value=self.condition_pool
            )
        )

    # another approach for sync to resolve an issue of AT REST API
    # query limitation(up to 500), especially when OR & IN conditions are used.
    # Split original get call into several small calls and
    # repeat with different api condition
    def get(self, results):

        field_ids = self.condition_pool
        batch_query_size = self.batch_query_size

        while field_ids:
            batch_condition = field_ids[:batch_query_size]
            del field_ids[:batch_query_size]
            self._replace_batch_conditions(self.client.conditions,
                                           batch_condition,
                                           self.condition_field_name)
            self.fetch_records(results)

        return results

    def _replace_batch_conditions(self, conditions, batch_condition,
                                  condition_field_name):
        for c in conditions:
            if c.field == condition_field_name:
                c.value = batch_condition

    @property
    def active_ids(self):
        raise NotImplementedError


class MultiConditionBatchQueryMixin(BatchQueryMixin):

    def _add_conditions(self):
        self.multi_conditions = {}
        for field_name, active_ids in self.active_ids.items():
            self.multi_conditions.update({
                field_name: A(
                    op='in',
                    field=field_name,
                    value=active_ids
                )
            })

    def get(self, results):

        for condition_field_name, condition in self.multi_conditions.items():
            field_ids = condition.value
            batch_query_size = self.batch_query_size
            idx_condition = self.client.add_condition(
                A(
                    op='in',
                    field=condition_field_name,
                    value=[]
                )
            )

            while field_ids:
                batch_condition = field_ids[:batch_query_size]
                del field_ids[:batch_query_size]
                self._replace_batch_conditions(self.client.conditions,
                                               batch_condition,
                                               condition_field_name)
                self.fetch_records(results)

            self.client.remove_condition(idx_condition)

        return results


class Synchronizer:
    lookup_key = 'id'
    last_updated_field = 'lastActivityDate'
    bulk_prune = True

    def __init__(self, full=False, *args, **kwargs):
        self.client = self.client_class(
            impersonation_resource=kwargs.get('impersonation_resource'),
        )
        self.full = full

    def set_relations(self, instance, json_data):
        for json_field, value in self.related_meta.items():
            model_class, field_name = value
            self._assign_relation(
                instance,
                json_data,
                json_field,
                model_class,
                field_name
            )

    @property
    def impersonation_resource(self):
        return self.client.impersonation_resource

    @staticmethod
    def _assign_null_relation(instance, model_field):
        """
        Set the FK to null, but handle issues like the FK being non-null.
        """
        try:
            setattr(instance, model_field, None)
        except ValueError:
            # The model_field may have been non-null.
            raise InvalidObjectException(
                "Unable to set field {} on {} to null, as it's required.".
                format(model_field, instance)
            )

    def _assign_relation(self, instance, json_data,
                         json_field, model_class, model_field):
        """
        Look up the given foreign relation, and set it to the given
        field on the instance.
        """
        uid = json_data.get(json_field)

        try:
            if uid is not None and uid != '':
                related_instance = model_class.objects.get(pk=uid)
                setattr(instance, model_field, related_instance)
            else:
                self._assign_null_relation(instance, model_field)
        except model_class.DoesNotExist:
            logger.warning(
                'Failed to find {} {} for {} {}.'.format(
                    json_field,
                    uid,
                    type(instance),
                    instance.id
                )
            )
            self._assign_null_relation(instance, model_field)

    def _instance_ids(self, filter_params=None):
        # self.lookup_key is used only for json_data. In DB, id is fixed for
        # primary key name
        db_lookup_key = 'id'
        if not filter_params:
            ids = self.model_class.objects.all().values_list(
                db_lookup_key, flat=True)
        else:
            ids = self.model_class.objects.filter(filter_params).values_list(
                db_lookup_key, flat=True)
        return set(ids)

    def get(self, results):
        return self.fetch_records(results)

    def fetch_records(self, results):
        """
        For all pages of results, save each page of results to the DB.
        """
        next_url = None
        while True:
            logger.info(
                'Fetching {} records'.format(
                    self.model_class.__bases__[0].__name__)
            )
            api_return = self.get_page(next_url)
            page = api_return.get("items")
            next_url = api_return.get("pageDetails").get("nextPageUrl")
            self.persist_page(page, results)

            if not next_url:
                break

        return results

    def persist_page(self, records, results):
        """Persist one page of records to DB."""
        for record in records:
            try:
                with transaction.atomic():
                    instance, result = self.update_or_create_instance(record)
                if result == CREATED:
                    results.created_count += 1
                elif result == UPDATED:
                    results.updated_count += 1
                else:
                    results.skipped_count += 1
            except InvalidObjectException as e:
                logger.warning('{}'.format(e))

            results.synced_ids.add(self.get_record_id(record))

        return results

    def get_record_id(self, record):
        return int(record[self.lookup_key])

    def get_page(self, next_url=None, *args, **kwargs):
        return self.client.get(next_url, *args, **kwargs)

    def get_single(self, instance_id):
        return self.client.get_single(instance_id)

    def _assign_field_data(self, instance, api_instance):
        raise NotImplementedError

    def _set_datetime_attribute(self, instance, attribute_name):
        if getattr(instance, attribute_name):
            setattr(instance,
                    attribute_name,
                    parse(getattr(instance, attribute_name))
                    )

    def fetch_sync_by_id(self, instance_id):
        api_instance = self.get_single(instance_id)
        instance, created = \
            self.update_or_create_instance(api_instance['item'])
        return instance

    def update_or_create_instance(self, api_instance):
        """
        Creates and returns an instance if it does not already exist.
        """
        result = None
        api_instance = self.remove_null_characters(api_instance)
        try:
            instance_pk = self.get_record_id(api_instance)
            instance = self.model_class.objects.get(pk=instance_pk)
        except self.model_class.DoesNotExist:
            instance = self.model_class()
            result = CREATED

        try:
            self._assign_field_data(instance, api_instance)

            # This will return the created instance, the updated instance, or
            # if the instance is skipped an unsaved copy of the instance.
            if result == CREATED:
                if self.model_class is models.TicketTracker:
                    instance.save(force_insert=True)
                else:
                    instance.save()
            elif instance.tracker.changed():
                instance.save()
                result = UPDATED
            else:
                result = SKIPPED
        except IntegrityError as e:
            # This can happen when multiple threads are creating the
            # same ticket at once.
            msg = "IntegrityError while attempting to create {}." \
                  " Error: {}".format(self.model_class, e)
            logger.error(msg)
            raise InvalidObjectException(msg)

        if result == CREATED:
            result_log = 'Created'
        elif result == UPDATED:
            result_log = 'Updated'
        else:
            result_log = 'Skipped'

        logger.info('{}: {} {}'.format(
            result_log,
            self.model_class.__bases__[0].__name__,
            instance
        ))

        return instance, result

    def prune_stale_records(self, initial_ids, synced_ids):
        """
        Delete records that existed when sync started but were
        not seen as we iterated through all records from REST API.
        """
        stale_ids = initial_ids - synced_ids
        deleted_count = 0
        if stale_ids:
            delete_qset = self.get_delete_qset(stale_ids)
            deleted_count = delete_qset.count()

            logger.info(
                'Removing {} stale records for model: {}'.format(
                    len(stale_ids), self.model_class.__bases__[0].__name__,
                )
            )
            if self.bulk_prune:
                delete_qset.delete()
            else:
                for instance in delete_qset:
                    try:
                        instance.delete()
                    except IntegrityError as e:
                        logger.exception(
                            'A database error occurred while attempting to '
                            'delete {} records. Error: {}'.format(
                                self.model_class.__bases__[0].__name__,
                                e.__cause__
                            )
                        )
        return deleted_count

    def get_delete_qset(self, stale_ids):
        return self.model_class.objects.filter(pk__in=stale_ids)

    def get_sync_job_qset(self):
        return models.SyncJob.objects.filter(
            entity_name=self.model_class.__bases__[0].__name__
        )

    def create(self, **kwargs):
        raise NotImplementedError()

    @log_sync_job
    def sync(self):
        sync_job_qset = self.get_sync_job_qset().filter(success=True)

        if sync_job_qset.count() > 1 and self.last_updated_field \
                and not self.full:
            last_sync_job_time = sync_job_qset.last().start_time.strftime(
                '%Y-%m-%dT%H:%M:%S.%fZ')
            self.client.add_condition(
                A(
                    field=self.last_updated_field,
                    value=last_sync_job_time,
                    op="gt"
                )
            )
        results = SyncResults()
        results = self.get(results)

        if self.full:
            # Set of IDs of all records prior to sync,
            # to find stale records for deletion.
            initial_ids = self._instance_ids()

            results.deleted_count = self.prune_stale_records(
                initial_ids, results.synced_ids
            )

        return results.created_count, results.updated_count, \
            results.skipped_count, results.deleted_count

    def remove_null_characters(self, json_data):
        for value in json_data:
            if isinstance(json_data.get(value), str):
                json_data[value] = json_data[value].replace('\x00', '')

        return json_data

    def _translate_fields_to_api_format(self, fields):
        """
        Converts the model field names to the API field names.
        """
        api_fields = {}
        for key, value in fields.items():
            try:
                api_fields[self.API_FIELD_NAMES[key]] = value
            except KeyError:
                # If the key is not in the API_FIELD_NAMES dict, continue
                pass
        return api_fields


class SyncRecordUDFMixin:

    def _assign_udf_data(self, instance, udfs):
        for item in udfs:
            try:
                name = item['name']
                value = item['value']

                udf = self.udf_class.objects.get(
                    name=name
                )

                instance.udf[str(udf.id)] = {
                    'name': name,
                    'value': value,
                    'label': udf.label,
                    'type': udf.type,
                    'is_picklist': udf.is_picklist
                }

                if value and udf.is_picklist:
                    # On a picklist item, the label is different from
                    # the name.
                    instance.udf[str(udf.id)]['label'] = \
                        udf.picklist[value]['label']

            except self.udf_class.MultipleObjectsReturned as e:
                # Shouldn't ever happen but just in case, log and continue
                logger.error(
                    'Multiple UDF records returned for 1 name: {}'.format(e))
            except self.udf_class.DoesNotExist as e:
                # Can happen if sync not 100% up to date, debug log and
                # continue
                logger.debug(
                    'No UDF records returned for name: {}'.format(e))
            except KeyError as e:
                # UDF has likely been updated but we don't have the
                # updated changes locally until the UDF class has been synced.
                logger.warning(
                    'KeyError when trying to access UDF '
                    'picklist label. {}'.format(e)
                )


class UDFSynchronizer(Synchronizer):
    lookup_key = 'name'
    model_class = None
    last_updated_field = None

    def get_record_id(self, record):
        try:
            return self.model_class.objects.\
                get(name=record[self.lookup_key]).id
        except self.model_class.DoesNotExist:
            return None

    def fetch_records(self, results):
        """
        For UDF, AT returns only 1 page of results; save the results to the DB.
        """
        next_url = self.client_class().get_api_url()
        logger.info(
            'Fetching {} records.'.format(self.model_class)
        )
        api_return = self.get_page(next_url)
        total_page = api_return.get("fields")
        if total_page:
            self.persist_page(total_page, results)

        return results

    def _assign_field_data(self, instance, object_data):

        instance.name = object_data.get('name')
        instance.label = object_data.get('label')
        instance.type = object_data.get('type')
        instance.is_picklist = object_data.get('isPickList')

        if instance.is_picklist:
            # Clear picklist to eliminate stale items
            instance.picklist = {}
            for item in object_data.get('picklistValues'):
                instance.picklist[item['value']] = {
                    'label': item['label'],
                    'is_default_value': item['isDefaultValue'],
                    'sort_order': item['sortOrder'],
                    'is_active': item['isActive'],
                    'is_system': item['isSystem'],
                }

        return instance


class TicketUDFSynchronizer(UDFSynchronizer):
    client_class = api.TicketsUDFAPIClient
    model_class = models.TicketUDFTracker


class TaskUDFSynchronizer(UDFSynchronizer):
    client_class = api.TasksUDFAPIClient
    model_class = models.TaskUDFTracker


class ProjectUDFSynchronizer(UDFSynchronizer):
    client_class = api.ProjectsUDFAPIClient
    model_class = models.ProjectUDFTracker


class CreateRecordMixin:

    def create(self, **kwargs):
        """
        Make a request to Autotask to create an entity.
        """
        instance = self.model_class()
        created_record_fields = self._translate_fields_to_api_format(kwargs)
        created_id = self.client.create(instance, **created_record_fields)

        # get_single retrieves the newly created entity info, which includes
        # generated/calculated fields from AT-side
        created_instance = self.get_single(created_id)

        return self.update_or_create_instance(created_instance['item'])


class UpdateRecordMixin:

    def update(self, instance, **kwargs):
        """
        Make a request to Autotask to update an entity.
        """
        updated_record_fields = self._translate_fields_to_api_format(kwargs)
        updated_id = self.client.update(instance, updated_record_fields)

        # get_single retrieves the updated entity info, to get the latest data
        updated_instance = self.get_single(updated_id['itemId'])

        return self.update_or_create_instance(updated_instance['item'])


class DeleteRecordMixin:

    def delete(self, instance, parent=None):
        """
        Make a request to Autotask to delete an entity.
        """
        try:
            deleted_instance_id = self.client.delete(instance, parent)
            if deleted_instance_id:
                instance.delete()
        except AutotaskRecordNotFoundError:
            # We can safely delete instance to sync
            instance.delete()


class ChildCreateRecordMixin:

    def create(self, parent, **kwargs):
        """
        Make a request to Autotask to create an entity.
        """
        instance = self.model_class()
        created_record_fields = self._translate_fields_to_api_format(kwargs)
        created_id = \
            self.client.create(instance, parent, **created_record_fields)

        # get_single retrieves the newly created entity info, which includes
        # generated/calculated fields from AT-side
        created_instance = self.get_single(created_id)

        return self.update_or_create_instance(created_instance['item'])


class ContactSynchronizer(Synchronizer):
    client_class = api.ContactsAPIClient
    model_class = models.ContactTracker

    related_meta = {
        'companyID': (models.Account, 'account'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client.add_condition(A(op='eq', field='isActive', value='true'))

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.first_name = json_data.get('firstName')
        instance.last_name = json_data.get('lastName')
        instance.email_address = json_data.get('emailAddress')
        instance.email_address2 = json_data.get('emailAddress2')
        instance.email_address3 = json_data.get('emailAddress3')
        instance.phone = json_data.get('phone')
        instance.alternate_phone = json_data.get('alternatePhone')
        instance.mobile_phone = json_data.get('mobilePhone')
        instance.extension = json_data.get('extension')

        self.set_relations(instance, json_data)

        return instance


class RoleSynchronizer(Synchronizer):
    client_class = api.RolesAPIClient
    model_class = models.RoleTracker
    last_updated_field = None

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.active = json_data.get('isActive')
        instance.name = json_data.get('name')
        instance.description = json_data.get('description')
        instance.hourly_factor = json_data.get('hourlyFactor')
        instance.hourly_rate = json_data.get('hourlyRate')
        instance.role_type = json_data.get('roleType')
        instance.system_role = json_data.get('isSystemRole')

        if instance.hourly_factor:
            instance.hourly_factor = \
                Decimal(str(round(instance.hourly_factor, 2)))
        if instance.hourly_rate:
            instance.hourly_rate = \
                Decimal(str(round(instance.hourly_rate, 2)))

        return instance


class DepartmentSynchronizer(Synchronizer):
    client_class = api.DepartmentsAPIClient
    model_class = models.DepartmentTracker
    last_updated_field = None

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data.get('name')
        instance.description = json_data.get('description')
        instance.number = json_data.get('number')

        return instance


class ResourceServiceDeskRoleSynchronizer(Synchronizer):
    client_class = api.ResourceServiceDeskRolesAPIClient
    model_class = models.ResourceServiceDeskRoleTracker
    last_updated_field = None

    related_meta = {
        'resourceID': (models.Resource, 'resource'),
        'roleID': (models.Role, 'role'),
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.active = json_data.get('isActive')
        instance.default = json_data.get('isDefault')

        self.set_relations(instance, json_data)

        return instance


class ResourceRoleDepartmentSynchronizer(Synchronizer):
    client_class = api.ResourceRoleDepartmentsAPIClient
    model_class = models.ResourceRoleDepartmentTracker
    last_updated_field = None

    related_meta = {
        'resourceID': (models.Resource, 'resource'),
        'roleID': (models.Role, 'role'),
        'departmentID': (models.Department, 'department'),
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.active = json_data.get('isActive')
        instance.default = json_data.get('isDefault')
        instance.department_lead = json_data.get('isDepartmentLead')

        self.set_relations(instance, json_data)

        return instance


class CompletedDateMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        request_settings = DjautotaskSettings().get_settings()
        completed_date = (timezone.now() - timezone.timedelta(
                hours=request_settings.get('keep_completed_hours')
            )).strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        self.client.add_condition(
            A(
                A(
                    op='gt',
                    field=self.completed_date_field,
                    value=completed_date
                ),
                A(op='noteq', field='status', value=models.Status.COMPLETE_ID),
                op='or'
            )
        )


class TicketSynchronizer(CreateRecordMixin,
                         UpdateRecordMixin,
                         SyncRecordUDFMixin,
                         CompletedDateMixin,
                         Synchronizer,
                         ParentSynchronizer):
    client_class = api.TicketsAPIClient
    model_class = models.TicketTracker
    udf_class = models.TicketUDF
    completed_date_field = 'completedDate'

    API_FIELD_NAMES = {
        'title': 'title',
        'description': 'description',
        'queue': 'queueID',
        'estimated_hours': 'estimatedHours',
        'due_date_time': 'dueDateTime',
        'status': 'status',
        'priority': 'priority',
        'category': 'ticketCategory',
        'billing_code': 'billingCodeID',
        'issue_type': 'issueType',
        'sub_issue_type': 'subIssueType',
        'project': 'projectID',
        'assigned_resource': 'assignedResourceID',
        'assigned_resource_role': 'assignedResourceRoleID',
        'account': 'companyID',
        'account_physical_location': 'companyLocationID',
        'contact': 'contactID',
        'contract': 'contractID',
    }

    def __init__(self, full=False, *args, **kwargs):
        settings = DjautotaskSettings().get_settings()
        self.queue_sync_filter = settings.get('queue_sync_filter')
        super().__init__(full, *args, **kwargs)
        self._add_conditions()

    def _add_conditions(self):
        condition_list = []
        for queue_id in self.queue_sync_filter:
            condition_list.append(
                A(
                    op='eq',
                    field='queueID',
                    value=queue_id,
                ),
            )

        if condition_list:
            self.client.add_condition(A(*condition_list, op="or"))

    related_meta = {
        'companyID': (models.Account, 'account'),
        'companyLocationID': (models.AccountPhysicalLocation,
                              'account_physical_location'),
        'status': (models.Status, 'status'),
        'assignedResourceID': (models.Resource, 'assigned_resource'),
        'priority': (models.Priority, 'priority'),
        'queueID': (models.Queue, 'queue'),
        'projectID': (models.Project, 'project'),
        'ticketCategory': (models.TicketCategory, 'category'),
        'ticketType': (models.TicketType, 'type'),
        'source': (models.Source, 'source'),
        'issueType': (models.IssueType, 'issue_type'),
        'subIssueType': (models.SubIssueType, 'sub_issue_type'),
        'assignedResourceRoleID': (models.Role, 'assigned_resource_role'),
        'billingCodeID': (models.BillingCode, 'billing_code'),
        'contractID': (models.Contract, 'contract'),
        'contactID': (models.Contact, 'contact'),
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.title = json_data['title']

        instance.ticket_number = json_data.get('ticketNumber')
        instance.description = json_data.get('description')
        instance.estimated_hours = json_data.get('estimatedHours')
        instance.service_level_agreement = \
            json_data.get('serviceLevelAgreementID')
        instance.service_level_agreement_has_been_met = \
            bool(json_data.get('serviceLevelAgreementHasBeenMet'))
        sla_paused = \
            json_data.get('serviceLevelAgreementPausedNextEventHours')
        instance.first_response_date_time = \
            json_data.get('firstResponseDateTime')
        instance.first_response_due_date_time = \
            json_data.get('firstResponseDueDateTime')
        instance.resolution_plan_date_time = \
            json_data.get('resolutionPlanDateTime')
        instance.resolution_plan_due_date_time = \
            json_data.get('resolutionPlanDueDateTime')
        instance.resolved_date_time = \
            json_data.get('resolvedDateTime')
        instance.resolved_due_date_time = \
            json_data.get('resolvedDueDateTime')
        instance.create_date = json_data.get('createDate')
        instance.due_date_time = json_data.get('dueDateTime')
        instance.completed_date = json_data.get('completedDate')
        instance.last_activity_date = json_data.get('lastActivityDate')

        self._set_datetime_attribute(instance, 'first_response_date_time')
        self._set_datetime_attribute(instance,
                                     'first_response_due_date_time')
        self._set_datetime_attribute(instance, 'resolution_plan_date_time')
        self._set_datetime_attribute(instance,
                                     'resolution_plan_due_date_time')
        self._set_datetime_attribute(instance, 'resolved_date_time')
        self._set_datetime_attribute(instance, 'resolved_due_date_time')
        self._set_datetime_attribute(instance, 'create_date')
        self._set_datetime_attribute(instance, 'due_date_time')
        self._set_datetime_attribute(instance, 'completed_date')
        self._set_datetime_attribute(instance, 'last_activity_date')

        udfs = json_data.get('userDefinedFields')

        # Refresh udf field to eliminate stale udfs
        instance.udf = dict()

        if len(udfs):
            self._assign_udf_data(instance, udfs)

        # It's important to set the SLA paused field to zero if it's not
        # present in the JSON data because we need to track when a ticket
        # moves from a paused SLA state to one where the SLA timer is running.
        if not sla_paused:
            sla_paused = 0

        instance.service_level_agreement_paused_next_event_hours = \
            Decimal(str(round(sla_paused, 2)))
        if instance.estimated_hours:
            instance.estimated_hours = \
                Decimal(str(round(instance.estimated_hours, 2)))

        self.set_relations(instance, json_data)
        return instance

    def fetch_sync_by_id(self, instance_id):
        instance = super().fetch_sync_by_id(instance_id)
        if not instance.status or \
                instance.status.id != models.Status.COMPLETE_ID:
            self.sync_related(instance)
        return instance

    def sync_related(self, instance):
        sync_classes = []
        query_params = ('ticketID', instance.id)

        sync_classes.append(
            (TicketNoteSynchronizer(), query_params)
        )
        sync_classes.append(
            (TimeEntrySynchronizer(), query_params)
        )
        sync_classes.append(
            (TicketSecondaryResourceSynchronizer(), query_params)
        )

        TicketChecklistItemsSynchronizer().sync_items(instance)

        self.sync_children(*sync_classes)

    def count(self, **kwargs):
        queue_id = kwargs['queue_id']

        tickets_api = self.client_class()

        # Create api condition to request tickets of queue ID
        tickets_api.add_condition(
            A(op="eq",
              field="queueID",
              value=f'{queue_id}'
              )
        )

        # Call Ticket API count method to make request
        return tickets_api.count(next_url=None)


class TaskSynchronizer(ChildCreateRecordMixin, SyncRecordUDFMixin,
                       CompletedDateMixin, BatchQueryMixin, Synchronizer):
    client_class = api.TasksAPIClient
    model_class = models.TaskTracker
    udf_class = models.TaskUDF
    completed_date_field = 'completedDateTime'
    condition_field_name = 'projectId'
    last_updated_field = 'lastActivityDateTime'

    related_meta = {
        'taskCategoryID': (models.TaskCategory, 'category'),
        'assignedResourceID': (models.Resource, 'assigned_resource'),
        'assignedResourceRoleID': (models.Role, 'assigned_resource_role'),
        'billingCodeID': (models.BillingCode, 'billing_code'),
        'departmentID': (models.Department, 'department'),
        'phaseID': (models.Phase, 'phase'),
        'projectID': (models.Project, 'project'),
        'priorityLabel': (models.Priority, 'priority'),
        'status': (models.Status, 'status'),
    }

    API_FIELD_NAMES = {
        'title': 'title',
        'description': 'description',
        'start_date': 'startDateTime',
        'end_date': 'endDateTime',
        'estimated_hours': 'estimatedHours',
        'remaining_hours': 'remainingHours',
        'status': 'status',
        'department': 'departmentID',
        'billing_code': 'billingCodeID',
        'priority': 'priorityLabel',
        'phase': 'phaseID',
        'assigned_resource': 'assignedResourceID',
        'assigned_resource_role': 'assignedResourceRoleID',
        'category': 'taskCategoryID',
        'project': 'projectID',
        'task_type': 'taskType',
    }

    @property
    def active_ids(self):
        active_projects = models.Project.objects.exclude(
            Q(status__is_active=False) |
            Q(status__id=models.ProjectStatus.COMPLETE_ID)
        ).values_list('id', flat=True).order_by(self.lookup_key)

        return active_projects

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.title = json_data['title']

        instance.number = json_data.get('taskNumber')
        instance.description = json_data.get('description')

        if instance.description:
            # Truncate the field to 8000 characters as per AT docs. Since we've
            # received descriptions greater than 8000 we'll truncate here
            # instead of catching the DataError that would be raised.
            # It is preferred to keep the DB schema in-line with the
            # AT specifications, even if they are wrong.
            instance.description = \
                instance.description[:models.Task.MAX_DESCRIPTION]
        instance.create_date = json_data.get('createDateTime')
        instance.completed_date = json_data.get('completedDateTime')
        instance.start_date = json_data.get('startDateTime')
        instance.end_date = json_data.get('endDateTime')
        instance.estimated_hours = json_data.get('estimatedHours')
        instance.remaining_hours = json_data.get('remainingHours')
        instance.last_activity_date = json_data.get('lastActivityDateTime')

        self._set_datetime_attribute(instance, 'create_date')
        self._set_datetime_attribute(instance, 'completed_date')
        self._set_datetime_attribute(instance, 'start_date')
        self._set_datetime_attribute(instance, 'end_date')
        self._set_datetime_attribute(instance, 'last_activity_date')

        udfs = json_data.get('userDefinedFields')

        # Refresh udf field to eliminate stale udfs
        instance.udf = dict()

        if len(udfs):
            self._assign_udf_data(instance, udfs)

        if instance.estimated_hours:
            instance.estimated_hours = \
                Decimal(str(round(instance.estimated_hours, 2)))
        if instance.remaining_hours:
            instance.remaining_hours = \
                Decimal(str(round(instance.remaining_hours, 2)))

        self.set_relations(instance, json_data)
        return instance

    def update(self, instance, **kwargs):
        """
        Make a request to Autotask to update an entity.
        """
        updated_record_fields = self._translate_fields_to_api_format(kwargs)

        updated_id = self.client.update(
            instance, instance.project, updated_record_fields)

        # get_single retrieves the updated entity info, to get the latest data
        updated_instance = self.get_single(updated_id['itemId'])

        return self.update_or_create_instance(updated_instance['item'])


class NoteSynchronizer(BatchQueryMixin, Synchronizer):

    API_FIELD_NAMES = {
        'title': 'title',
        'description': 'description',
        'note_type': 'noteType',
        'publish': 'publish',
        'created_by_contact_id': 'createdByContactID',
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client.add_condition(
            A(op='noteq', field='noteType',
              value=models.NoteType.WORKFLOW_RULE_NOTE_ID)
        )

    def _assign_field_data(self, instance, object_data):

        instance.id = object_data['id']
        instance.title = object_data.get('title')
        instance.description = object_data.get('description')
        instance.create_date_time = object_data.get('createDateTime')
        instance.last_activity_date = object_data.get('lastActivityDate')
        instance.publish = str(object_data.get('publish'))

        self._set_datetime_attribute(instance, 'create_date_time')
        self._set_datetime_attribute(instance, 'last_activity_date')

        if instance.description:
            # Autotask docs say the max description length is 3200
            # characters but we've seen descriptions that are longer than that.
            # So truncate the field to 3200 characters just in case.
            instance.description = instance.description[:3200]

        self.set_relations(instance, object_data)

        return instance


class TicketNoteSynchronizer(ChildCreateRecordMixin,
                             NoteSynchronizer,
                             ChildSynchronizer):
    client_class = api.TicketNotesAPIClient
    model_class = models.TicketNoteTracker
    condition_field_name = 'ticketID'
    related_instance_name = 'ticket'

    related_meta = {
        'noteType': (models.NoteType, 'note_type'),
        'creatorResourceID': (models.Resource, 'creator_resource'),
        'ticketID': (models.Ticket, 'ticket'),
        'createdByContactID': (models.Contact, 'created_by_contact'),
    }

    @property
    def active_ids(self):
        active_ids = models.Ticket.objects.all().\
            values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids

    def _build_children_conditions(self, query_params):
        parent_field, parent_id = query_params
        self.client.clear_conditions()
        self.client.add_condition(
            A(
                A(op='noteq', field='noteType',
                  value=models.NoteType.WORKFLOW_RULE_NOTE_ID),
                A(
                    op='eq',
                    field=parent_field,
                    value=parent_id
                ),
                op='and'
            )
        )

    def create(self, parent, **kwargs):
        """
        Make a request to Autotask to create a Note.
        """

        return super().create(parent, **kwargs)


class TaskNoteSynchronizer(CreateRecordMixin, NoteSynchronizer):
    client_class = api.TaskNotesAPIClient
    model_class = models.TaskNoteTracker
    condition_field_name = 'taskID'
    related_instance_name = 'task'

    related_meta = {
        'noteType': (models.NoteType, 'note_type'),
        'creatorResourceID': (models.Resource, 'creator_resource'),
        'taskID': (models.Task, 'task'),
        'createdByContactID': (models.Contact, 'created_by_contact'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        super().API_FIELD_NAMES.update({
            'task': 'taskID',
        })

    @property
    def active_ids(self):
        active_ids = models.Task.objects.exclude(
            status=models.Status.COMPLETE_ID
        ).values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids


class TimeEntrySynchronizer(CreateRecordMixin,
                            MultiConditionBatchQueryMixin,
                            Synchronizer,
                            ChildSynchronizer):
    client_class = api.TimeEntriesAPIClient
    model_class = models.TimeEntryTracker
    last_updated_field = 'lastModifiedDateTime'

    related_meta = {
        'resourceID': (models.Resource, 'resource'),
        'ticketID': (models.Ticket, 'ticket'),
        'taskID': (models.Task, 'task'),
        'timeEntryType': (models.TaskTypeLink, 'type'),
        'billingCodeID': (models.BillingCode, 'billing_code'),
        'roleID': (models.Role, 'role'),
        'contractID': (models.Contract, 'contract'),
    }
    API_FIELD_NAMES = {
        'ticket': 'ticketID',
        'task': 'taskID',
        'date_worked': 'dateWorked',
        'start_date_time': 'startDateTime',
        'end_date_time': 'endDateTime',
        'summary_notes': 'summaryNotes',
        'internal_notes': 'internalNotes',
        'hours_worked': 'hoursWorked',
        'offset_hours': 'offsetHours',
        'role': 'roleID',
        'resource': 'resourceID',
        'billing_code': 'billingCodeID',
        'contract': 'contractID',
        'show_on_invoice': 'showOnInvoice'
    }

    @property
    def active_ids(self):
        active_tickets = models.Ticket.objects.all().values_list(
            'id', flat=True).order_by(self.lookup_key)
        active_tasks = models.Task.objects.all().values_list(
            'id', flat=True).order_by(self.lookup_key)

        active_ids = dict()
        if len(active_tickets):
            active_ids.update({'ticketID': list(active_tickets)})
        if len(active_tasks):
            active_ids.update({'taskID': list(active_tasks)})
        return active_ids

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.date_worked = object_data.get('dateWorked')
        instance.start_date_time = object_data.get('startDateTime')
        instance.end_date_time = object_data.get('endDateTime')
        instance.summary_notes = object_data.get('summaryNotes')
        instance.internal_notes = object_data.get('internalNotes')
        instance.non_billable = object_data.get('isNonBillable')
        instance.hours_worked = object_data.get('hoursWorked')
        instance.hours_to_bill = object_data.get('hoursToBill')
        instance.offset_hours = object_data.get('offsetHours')

        self._set_datetime_attribute(instance, 'date_worked')
        self._set_datetime_attribute(instance, 'start_date_time')
        self._set_datetime_attribute(instance, 'end_date_time')

        if instance.hours_worked:
            instance.hours_worked = \
                Decimal(str(round(instance.hours_worked, 4)))
        if instance.hours_to_bill:
            instance.hours_to_bill = \
                Decimal(str(round(instance.hours_to_bill, 4)))
        if instance.offset_hours:
            instance.offset_hours = \
                Decimal(str(round(instance.offset_hours, 4)))

        self.set_relations(instance, object_data)
        return instance


class SecondaryResourceSynchronizer(ChildCreateRecordMixin, DeleteRecordMixin,
                                    BatchQueryMixin, Synchronizer):

    def delete(self, instances, parent=None):
        """
        Make a request to Autotask to delete a SecondaryResource.
        """
        for instance in instances:
            super().delete(instance, parent)

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        self.set_relations(instance, object_data)
        return instance

    @property
    def active_ids(self):
        active_ids = self.related_instance_model.objects.all(). \
            values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids


class TicketSecondaryResourceSynchronizer(SecondaryResourceSynchronizer,
                                          ChildSynchronizer):
    client_class = api.TicketSecondaryResourcesAPIClient
    model_class = models.TicketSecondaryResourceTracker
    last_updated_field = None
    condition_field_name = 'ticketID'
    related_instance_name = 'ticket'
    related_instance_model = models.Ticket

    related_meta = {
        'resourceID': (models.Resource, 'resource'),
        'roleID': (models.Role, 'role'),
        'ticketID': (models.Ticket, 'ticket'),
    }
    API_FIELD_NAMES = {
        'resource': 'resourceID',
        'role': 'roleID',
        'ticket': 'ticketID',
    }


class TaskSecondaryResourceSynchronizer(SecondaryResourceSynchronizer):
    client_class = api.TaskSecondaryResourcesAPIClient
    model_class = models.TaskSecondaryResourceTracker
    last_updated_field = None
    condition_field_name = 'taskID'
    related_instance_name = 'task'
    related_instance_model = models.Task

    related_meta = {
        'resourceID': (models.Resource, 'resource'),
        'roleID': (models.Role, 'role'),
        'taskID': (models.Task, 'task'),
    }
    API_FIELD_NAMES = {
        'resource': 'resourceID',
        'role': 'roleID',
        'task': 'taskID',
    }


class ProjectSynchronizer(CreateRecordMixin, UpdateRecordMixin,
                          SyncRecordUDFMixin, Synchronizer,
                          ParentSynchronizer, CompletedDateMixin):
    client_class = api.ProjectsAPIClient
    model_class = models.ProjectTracker
    udf_class = models.ProjectUDF
    last_updated_field = 'lastActivityDateTime'
    completed_date_field = 'completedDateTime'

    related_meta = {
        'projectLeadResourceID': (models.Resource, 'project_lead_resource'),
        'companyID': (models.Account, 'account'),
        'status': (models.ProjectStatus, 'status'),
        'projectType': (models.ProjectType, 'type'),
        'contractID': (models.Contract, 'contract'),
        'department': (models.Department, 'department'),
    }

    API_FIELD_NAMES = {
        'name': 'projectName',
        'status': 'status',
        'description': 'description',
        'start_date': 'startDateTime',
        'end_date': 'endDateTime',
        'type': 'projectType',
        'project_lead_resource': 'projectLeadResourceID',
        'department': 'department',
        'status_detail': 'statusDetail',
        'account': 'companyID'
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client.add_condition(
            A(
                A(
                    op='noteq',
                    field='status',
                    value=models.ProjectStatus.COMPLETE_ID
                ),
                A(
                    op='in',
                    field='status',
                    value=list(self.get_active_ids())
                ),
                op="and"
            )
        )

    def get_active_ids(self):
        active_project_statuses = models.ProjectStatus.objects.exclude(
            is_active=False).values_list('id', flat=True).order_by(
            self.lookup_key)

        return active_project_statuses

    def _assign_field_data(self, instance, object_data):

        instance.id = object_data['id']
        instance.name = object_data.get('projectName')
        instance.number = object_data.get('projectNumber')
        instance.description = object_data.get('description')
        instance.actual_hours = object_data.get('actualHours')
        instance.completed_percentage = object_data.get('completedPercentage')
        instance.duration = object_data.get('duration')
        instance.estimated_time = object_data.get('estimatedTime')
        instance.status_detail = object_data.get('statusDetail')

        instance.completed_date = object_data.get('completedDateTime')
        instance.end_date = object_data.get('endDateTime')
        instance.start_date = object_data.get('startDateTime')
        instance.last_activity_date_time = \
            object_data.get('lastActivityDateTime')

        self._set_datetime_attribute(instance, 'completed_date')
        self._set_datetime_attribute(instance, 'end_date')
        self._set_datetime_attribute(instance, 'start_date')
        self._set_datetime_attribute(instance, 'last_activity_date_time')

        if instance.completed_date:
            instance.completed_date = instance.completed_date.date()
        if instance.end_date:
            instance.end_date = instance.end_date.date()
        if instance.start_date:
            instance.start_date = instance.start_date.date()

        udfs = object_data.get('userDefinedFields')

        # Refresh udf field to eliminate stale udfs
        instance.udf = dict()

        if len(udfs):
            self._assign_udf_data(instance, udfs)

        if instance.estimated_time:
            instance.estimated_time = \
                Decimal(str(round(instance.estimated_time, 2)))
        if instance.actual_hours:
            instance.actual_hours = \
                Decimal(str(round(instance.actual_hours, 2)))

        if instance.description:
            # Autotask docs say the max description length is 2000
            # characters but we've seen descriptions that are longer than that.
            # So truncate the field to 2000 characters just in case.
            instance.description = instance.description[:2000]

        self.set_relations(instance, object_data)

        return instance


class TicketCategorySynchronizer(Synchronizer):
    client_class = api.TicketCategoriesAPIClient
    model_class = models.TicketCategoryTracker
    last_updated_field = None

    related_meta = {
        'displayColorRGB': (models.DisplayColor, 'display_color')
    }

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.name = json_data.get('name')
        instance.active = json_data.get('isActive')

        self.set_relations(instance, json_data)

        return instance


class TaskPredecessorSynchronizer(BatchQueryMixin, Synchronizer):
    client_class = api.TaskPredecessorsAPIClient
    model_class = models.TaskPredecessorTracker
    condition_field_name = 'predecessorTaskID'
    last_updated_field = None

    related_meta = {
        'predecessorTaskID': (models.Task, 'predecessor_task'),
        'successorTaskID': (models.Task, 'successor_task'),
    }

    @property
    def active_ids(self):
        active_tasks = models.Task.objects.exclude(
            Q(status__is_active=False) |
            Q(status__id=models.Status.COMPLETE_ID)
        ).values_list('id', flat=True).order_by(self.lookup_key)

        return active_tasks

    def _assign_field_data(self, instance, json_data):
        instance.id = json_data['id']
        instance.lag_days = json_data.get('lagDays')

        self.set_relations(instance, json_data)

        return instance


class ServiceCallSynchronizer(
        CreateRecordMixin, BatchQueryMixin, Synchronizer):

    client_class = api.ServiceCallsAPIClient
    model_class = models.ServiceCallTracker
    condition_field_name = 'companyID'
    last_updated_field = 'lastModifiedDateTime'

    related_meta = {
        'companyID': (models.Account, 'account'),
        'companyLocationID':
            (models.AccountPhysicalLocation, 'location'),
        'status': (models.ServiceCallStatus, 'status'),
        'creatorResourceID': (models.Resource, 'creator_resource'),
        'canceledByResourceID': (models.Resource, 'canceled_by_resource')
    }

    API_FIELD_NAMES = {
        'description': 'description',
        'duration': 'duration',
        'complete': 'isComplete',
        'create_date_time': 'createDateTime',
        'start_date_time': 'startDateTime',
        'end_date_time': 'endDateTime',
        'canceled_date_time': 'canceledDateTime',
        'last_modified_date_time': 'lastModifiedDateTime',
        'account': 'companyID',
        'location': 'companyLocationID',
        'status': 'status',
        'creator_resource': 'creatorResourceID',
        'canceled_by_resource': 'canceledByResourceID',
    }

    @property
    def active_ids(self):
        active_ids = models.Account.objects.filter(
            active=True
        ).values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.description = object_data.get('description')
        instance.duration = Decimal(str(object_data.get('duration')))
        instance.complete = object_data.get('isComplete')
        instance.create_date_time = object_data.get('createDateTime')
        instance.start_date_time = object_data.get('startDateTime')
        instance.end_date_time = object_data.get('endDateTime')
        instance.canceled_date_time = object_data.get('canceledDateTime')
        instance.last_modified_date_time = \
            object_data.get('lastModifiedDateTime')

        self._set_datetime_attribute(instance, 'create_date_time')
        self._set_datetime_attribute(instance, 'start_date_time')
        self._set_datetime_attribute(instance, 'end_date_time')
        self._set_datetime_attribute(instance, 'canceled_date_time')
        self._set_datetime_attribute(instance, 'last_modified_date_time')

        self.set_relations(instance, object_data)

        return instance


class ServiceCallTicketSynchronizer(
        ChildCreateRecordMixin, BatchQueryMixin, Synchronizer):

    client_class = api.ServiceCallTicketsAPIClient
    model_class = models.ServiceCallTicketTracker
    condition_field_name = 'ticketID'
    last_updated_field = None

    related_meta = {
        'serviceCallID': (models.ServiceCall, 'service_call'),
        'ticketID': (models.Ticket, 'ticket')
    }
    API_FIELD_NAMES = {
        'service_call': 'serviceCallID',
        'ticket': 'ticketID',
    }

    @property
    def active_ids(self):
        active_ids = models.Ticket.objects.all().\
            values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        self.set_relations(instance, object_data)

        return instance


class ServiceCallTaskSynchronizer(
        ChildCreateRecordMixin, BatchQueryMixin, Synchronizer):

    client_class = api.ServiceCallTasksAPIClient
    model_class = models.ServiceCallTaskTracker
    condition_field_name = 'taskID'
    last_updated_field = None

    related_meta = {
        'serviceCallID': (models.ServiceCall, 'service_call'),
        'taskID': (models.Task, 'task')
    }
    API_FIELD_NAMES = {
        'service_call': 'serviceCallID',
        'task': 'taskID',
    }

    @property
    def active_ids(self):
        active_ids = models.Task.objects.exclude(
            status=models.Status.COMPLETE_ID
        ).values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        self.set_relations(instance, object_data)

        return instance


class ServiceCallTicketResourceSynchronizer(
        ChildCreateRecordMixin, BatchQueryMixin, Synchronizer):

    client_class = api.ServiceCallTicketResourcesAPIClient
    model_class = models.ServiceCallTicketResourceTracker
    condition_field_name = 'serviceCallTicketID'
    last_updated_field = None

    related_meta = {
        'serviceCallTicketID':
            (models.ServiceCallTicket, 'service_call_ticket'),
        'resourceID': (models.Resource, 'resource')
    }
    API_FIELD_NAMES = {
        'resource': 'resourceID',
        'service_call_ticket': 'serviceCallTicketID',
    }

    @property
    def active_ids(self):
        active_ids = models.ServiceCallTicket.objects.all().\
            values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        self.set_relations(instance, object_data)

        return instance


class ServiceCallTaskResourceSynchronizer(
        ChildCreateRecordMixin, BatchQueryMixin, Synchronizer):

    client_class = api.ServiceCallTaskResourcesAPIClient
    model_class = models.ServiceCallTaskResourceTracker
    condition_field_name = 'serviceCallTaskID'
    last_updated_field = None

    related_meta = {
        'serviceCallTaskID':
            (models.ServiceCallTask, 'service_call_task'),
        'resourceID': (models.Resource, 'resource')
    }
    API_FIELD_NAMES = {
        'resource': 'resourceID',
        'service_call_task': 'serviceCallTaskID',
    }

    @property
    def active_ids(self):
        active_ids = models.ServiceCallTask.objects.all().\
            values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        self.set_relations(instance, object_data)

        return instance


class AttachmentSynchronizer(Synchronizer):
    client_class = api.AttachmentInfoAPIClient

    def __init__(self, record_type=None, field=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.type = record_type
        self.field = field

        if kwargs.get('object_id'):
            self.client.add_condition(
                A(op='eq', field=self.field, value=kwargs.get('object_id'))
            )

    def download_attachment(self, object_id, attachment_id, path):
        response = self.client.get_attachment(object_id,
                                              attachment_id,
                                              self.type)

        filename = response.get('fullPath')
        unique_filename = f'{attachment_id}-{filename}'

        logger.debug(f'Writing attachment {unique_filename} to {path}')

        file_path = os.path.join(path, f'{unique_filename}')
        # Set permissions on file before creating and writing to it.
        previous_umask = os.umask(FILE_UMASK)
        data = response.get('data')
        decoded_data = base64.b64decode(data)

        with open(file_path, 'wb') as f:
            f.write(decoded_data)
        os.umask(previous_umask)

        return unique_filename

    def get_count(self):
        return self.client.count()


class BillingCodeSynchronizer(Synchronizer):
    client_class = api.BillingCodesAPIClient
    model_class = models.BillingCodeTracker
    last_updated_field = None

    related_meta = {
        'useType': (models.UseType, 'use_type'),
        'billingCodeType': (models.BillingCodeType, 'billing_code_type')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client.add_condition(A(op='eq', field='isActive', value=True))

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.name = object_data.get('name')
        instance.description = object_data.get('description')
        instance.active = object_data.get('isActive')

        self.set_relations(instance, object_data)

        return instance


class ContractSynchronizer(Synchronizer):
    client_class = api.ContractsAPIClient
    model_class = models.ContractTracker
    last_updated_field = None

    related_meta = {
        'companyID': (models.Account, 'account')
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client.add_condition(A(op='eq', field='status', value='1'))

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.name = object_data.get('contractName')
        instance.number = object_data.get('contractNumber')
        instance.status = str(object_data.get('status'))
        instance.contract_exclusion_set_id = \
            object_data.get('contractExclusionSetID')

        self.set_relations(instance, object_data)

        return instance


class AccountPhysicalLocationSynchronizer(BatchQueryMixin, Synchronizer):
    client_class = api.AccountPhysicalLocationsAPIClient
    model_class = models.AccountPhysicalLocationTracker
    condition_field_name = 'companyID'
    last_updated_field = None

    related_meta = {
        'companyID': (models.Account, 'account'),
    }

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.name = object_data.get('name')
        instance.active = object_data.get('isActive')
        instance.primary = object_data.get('isPrimary')
        self.set_relations(instance, object_data)

        return instance

    @property
    def active_ids(self):
        active_ids = models.Account.objects.all().\
            values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids


class ResourceSynchronizer(Synchronizer):
    client_class = api.ResourcesAPIClient
    model_class = models.ResourceTracker
    last_updated_field = None
    bulk_prune = False

    related_meta = {
        'licenseType': (models.LicenseType, 'license_type'),
        'defaultServiceDeskRoleID': (models.Role, 'default_service_desk_role'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client.add_condition(A(op='eq', field='isActive', value=True))

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.user_name = object_data.get('userName')
        instance.email = object_data.get('email')
        instance.first_name = object_data.get('firstName')
        instance.last_name = object_data.get('lastName')
        instance.active = object_data.get('isActive')

        self.set_relations(instance, object_data)

        return instance


class AccountSynchronizer(Synchronizer):
    client_class = api.AccountsAPIClient
    model_class = models.AccountTracker

    related_meta = {
        'companyType': (models.AccountType, 'type'),
        'parentCompanyID': (models.Account, 'parent_account'),
        'ownerResourceID': (models.Resource, 'owner_resource'),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client.add_condition(A(op='eq', field='isActive', value=True))

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.name = object_data.get('companyName')
        instance.number = object_data.get('companyNumber')
        instance.active = object_data.get('isActive')
        instance.last_activity_date = object_data.get('lastActivityDate')
        instance.phone = object_data.get('phone')

        self._set_datetime_attribute(instance, 'last_activity_date')

        self.set_relations(instance, object_data)

        return instance


class PhaseSynchronizer(Synchronizer):
    client_class = api.PhasesAPIClient
    model_class = models.PhaseTracker
    last_updated_field = 'lastActivityDateTime'

    related_meta = {
        'projectID': (models.Project, 'project'),
        'parentPhaseID': (models.Phase, 'parent_phase'),
    }

    def _assign_field_data(self, instance, object_data):

        instance.id = object_data['id']
        instance.title = object_data.get('title')
        instance.number = object_data.get('phaseNumber')
        instance.description = object_data.get('description')
        instance.start_date = object_data.get('startDate')
        instance.due_date = object_data.get('dueDate')
        instance.last_activity_date = object_data.get('lastActivityDateTime')

        estimated_hours = object_data.get('estimatedHours')
        instance.estimated_hours = Decimal(str(estimated_hours)) \
            if estimated_hours is not None else None

        self._set_datetime_attribute(instance, 'start_date')
        self._set_datetime_attribute(instance, 'due_date')
        self._set_datetime_attribute(instance, 'last_activity_date')

        self.set_relations(instance, object_data)

        return instance


class PicklistSynchronizer(Synchronizer):
    lookup_name = None
    lookup_key = 'value'
    last_updated_field = None

    def fetch_records(self, results):
        logger.info(
            'Fetching {} records'.format(
                self.model_class.__bases__[0].__name__)
        )
        api_return = self.get_page()
        fields = api_return.get("fields")
        page = None

        if fields:
            for field in fields:
                try:
                    if field.get("name") == self.lookup_name:
                        page = field["picklistValues"]
                        break
                except (AttributeError, NameError):
                    pass

        if page:
            self.persist_page(page, results)

        return results

    def _assign_field_data(self, instance, json_data):
        instance.id = int(json_data[self.lookup_key])
        instance.label = json_data.get('label')
        instance.is_default_value = json_data.get('isDefaultValue')
        instance.sort_order = json_data.get('sortOrder')
        instance.is_active = json_data.get('isActive')
        instance.is_system = json_data.get('isSystem')

        if hasattr(self, 'related_meta'):
            self.set_relations(instance, json_data)
        return instance


class CompanyAlertSynchronizer(BatchQueryMixin, Synchronizer):
    client_class = api.CompanyAlertAPIClient
    model_class = models.CompanyAlertTracker
    condition_field_name = 'companyID'
    last_updated_field = None

    related_meta = {
        'companyID': (models.Account, 'account'),
    }

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.alert_text = object_data.get('alertText')
        instance.alert_type = object_data.get('alertTypeID')

        self.set_relations(instance, object_data)

        return instance

    @property
    def active_ids(self):
        active_ids = models.Account.objects.all().\
            values_list('id', flat=True).order_by(self.lookup_key)

        return active_ids


class ContractExclusionSetSynchronizer(Synchronizer):
    client_class = api.ContractExclusionSetAPIClient
    model_class = models.ContractExclusionSetTracker
    last_updated_field = None

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        instance.name = object_data['name']
        instance.is_active = object_data['isActive']
        instance.description = object_data['description']

        return instance


class ContractExcludedWorkTypeSynchronizer(BatchQueryMixin, Synchronizer):
    client_class = api.ContractsExcludedWorkTypeAPIClient
    model_class = models.ContractExcludedWorkTypeTracker
    condition_field_name = 'contractExclusionSetID'
    last_updated_field = None

    related_meta = {
        'contractExclusionSetID':
            (models.ContractExclusionSet, 'contract_exclusion_set'),
        'excludedWorkTypeID': (models.BillingCode, 'excluded_work_type')
    }

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        self.set_relations(instance, object_data)

        return instance

    @property
    def active_ids(self):
        active_ids = models.ContractExclusionSet.objects.exclude(
            is_active=False
        ).values_list(
            'id', flat=True
        )
        return active_ids


class ContractExcludedRoleSynchronizer(BatchQueryMixin, Synchronizer):
    client_class = api.ContractsExcludedRoleAPIClient
    model_class = models.ContractExcludedRoleTracker
    condition_field_name = 'contractExclusionSetID'
    last_updated_field = None

    related_meta = {
        'contractExclusionSetID':
            (models.ContractExclusionSet, 'contract_exclusion_set'),
        'excludedRoleID': (models.Role, 'excluded_role')
    }

    def _assign_field_data(self, instance, object_data):
        instance.id = object_data['id']
        self.set_relations(instance, object_data)

        return instance

    @property
    def active_ids(self):
        active_ids = models.ContractExclusionSet.objects.exclude(
            is_active=False
        ).values_list(
            'id', flat=True
        )
        return active_ids


class NoteTypeSynchronizer(PicklistSynchronizer):
    # Ticket note types are including task note types, and there are other
    # note types currently not used. e.g. project note types
    client_class = api.NoteTypesAPIClient
    model_class = models.NoteTypeTracker
    lookup_name = 'noteType'


class LicenseTypeSynchronizer(PicklistSynchronizer):
    client_class = api.LicenseTypesAPIClient
    model_class = models.LicenseTypeTracker
    lookup_name = 'licenseType'


class UseTypeSynchronizer(PicklistSynchronizer):
    client_class = api.UseTypesAPIClient
    model_class = models.UseTypeTracker
    lookup_name = 'useType'


class BillingCodeTypeSynchronizer(PicklistSynchronizer):
    client_class = api.BillingCodeTypeAPIClient
    model_class = models.BillingCodeTypeTracker
    lookup_name = 'billingCodeType'


class TaskTypeLinkSynchronizer(PicklistSynchronizer):
    client_class = api.TaskTypeLinksAPIClient
    model_class = models.TaskTypeLinkTracker
    lookup_name = 'timeEntryType'


class AccountTypeSynchronizer(PicklistSynchronizer):
    client_class = api.AccountTypesAPIClient
    model_class = models.AccountTypeTracker
    lookup_name = 'companyType'


class DisplayColorSynchronizer(PicklistSynchronizer):
    client_class = api.TicketCategoryPicklistAPIClient
    model_class = models.DisplayColorTracker
    lookup_name = 'displayColorRgb'


class ServiceCallStatusSynchronizer(PicklistSynchronizer):
    client_class = api.ServiceCallStatusPicklistAPIClient
    model_class = models.ServiceCallStatusTracker
    lookup_name = 'status'


class TicketPicklistSynchronizer(PicklistSynchronizer):
    client_class = api.TicketPicklistAPIClient


class StatusSynchronizer(TicketPicklistSynchronizer):
    model_class = models.StatusTracker
    lookup_name = 'status'


class PrioritySynchronizer(TicketPicklistSynchronizer):
    model_class = models.PriorityTracker
    lookup_name = 'priority'


class QueueSynchronizer(TicketPicklistSynchronizer):
    model_class = models.QueueTracker
    lookup_name = 'queueID'


class SourceSynchronizer(TicketPicklistSynchronizer):
    model_class = models.SourceTracker
    lookup_name = 'source'


class IssueTypeSynchronizer(TicketPicklistSynchronizer):
    model_class = models.IssueTypeTracker
    lookup_name = 'issueType'


class SubIssueTypeSynchronizer(TicketPicklistSynchronizer):
    model_class = models.SubIssueTypeTracker
    lookup_name = 'subIssueType'

    related_meta = {
        'parentValue': (models.IssueType, 'parent_value'),
    }


class TicketTypeSynchronizer(TicketPicklistSynchronizer):
    model_class = models.TicketTypeTracker
    lookup_name = 'ticketType'


class TaskPicklistSynchronizer(PicklistSynchronizer):
    client_class = api.TaskPicklistAPIClient


class TaskCategorySynchronizer(TaskPicklistSynchronizer):
    model_class = models.TaskCategoryTracker
    lookup_name = 'taskCategoryID'


class TaskTypeSynchronizer(TaskPicklistSynchronizer):
    model_class = models.TaskTypeTracker
    lookup_name = 'taskType'


class ProjectPicklistSynchronizer(PicklistSynchronizer):
    client_class = api.ProjectPicklistAPIClient


class ProjectStatusSynchronizer(ProjectPicklistSynchronizer):
    model_class = models.ProjectStatusTracker
    lookup_name = 'status'


class ProjectTypeSynchronizer(ProjectPicklistSynchronizer):
    model_class = models.ProjectTypeTracker
    lookup_name = 'projectType'


class ProjectNoteTypeSynchronizer(PicklistSynchronizer):
    client_class = api.ProjectNoteTypesAPIClient
    model_class = models.ProjectNoteTypeTracker
    lookup_name = 'noteType'


class ProjectNotesSynchronizer(Synchronizer):
    client_class = api.ProjectNotesAPIClient

    def __init__(self, project_id=None, field=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if project_id:
            self.client.add_condition(
                A(op='eq', field=field, value=project_id)
            )


###################################################################
# Dummy Synchronizers                                             #
###################################################################


class DummySynchronizer:
    # Use FIELDS to list fields we submit to create or update a record, used
    # as a kind of validation method and way to link the snake_case field
    # names to their camelCase api names
    FIELDS = {}

    def __init__(self):
        self.client = self.client_class()

    def get(self, parent=None, conditions=None):

        records = []
        next_url = None
        conditions = conditions or []

        while True:
            logger.info(
                'Fetching {} records'.format(self.record_name)
            )

            api_return = self._get_page(
                next_url, conditions, parent=parent)
            records += api_return.get("items")
            next_url = api_return.get("pageDetails").get("nextPageUrl")

            if not next_url:
                break

        inverted = {v: k for k, v in self.FIELDS.items()}
        formatted_records = []

        # Convert the results from camelCase back to snake_case
        for record in records:
            converted = {}
            for k, v in record.items():
                if inverted.get(k, None):
                    converted[inverted[k]] = v
            formatted_records.append(converted)

        return formatted_records

    def update(self, parent=None, **kwargs):
        raise NotImplementedError

    def create(self, parent=None, **kwargs):
        raise NotImplementedError

    def delete(self, parent=None, **kwargs):
        raise NotImplementedError

    def _get_page(self, next_url, conditions, parent=None):
        raise NotImplementedError

    def _format_record(self, **kwargs):
        # Convert records from snake_case to camelCase
        record = {}
        for key, value in kwargs.items():
            # Only consider fields of the record, discard anything else
            if key in self.FIELDS.keys():
                record[self.FIELDS[key]] = value

        return record


class TicketChecklistItemsSynchronizer(DummySynchronizer):
    FIELDS = {
        'id': 'id',
        'ticket': 'ticketID',
        'item_name': 'itemName',
        'important': 'isImportant',
        'completed': 'isCompleted',
        'position': 'position',
        'resource': 'completedByResourceID',
        'completed_time': 'completedDateTime',
    }
    client_class = api.TicketChecklistItemsAPIClient
    record_name = "TicketChecklistItems"

    # https://webservices2.autotask.net/atservicesrest/V1.0/Tickets/9980/ChecklistItems
    def update(self, parent=None, **kwargs):
        record = self._format_record(**kwargs)

        return self.client.update(parent, **record)

    def create(self, parent=None, **kwargs):
        record = self._format_record(**kwargs)

        return self.client.create(parent, **record)

    def delete(self, parent=None, **kwargs):
        return self.client.delete(parent, **kwargs)

    def sync(self):
        ticket_qs = models.Ticket.objects.all().order_by('id')

        for ticket in ticket_qs:
            self.sync_items(ticket)

    def sync_items(self, instance):
        tasks = self.get(parent=instance.id)
        instance.checklist_total = len(tasks)
        instance.checklist_completed = sum(task['completed'] for task in tasks)

        instance.save()

    def _get_page(self, next_url, conditions, parent=None):
        if parent:
            self.client.add_condition(
                A(op='eq', field='ticketID', value=parent))

        for condition in conditions:
            self.client.add_condition(condition)

        return self.client.get(next_url, conditions)



import datetime
import decimal
import json
import logging
from json import JSONDecodeError

import pytz
import requests
from django.apps import apps
from django.conf import settings
from django.core.cache import cache
from django.db import models
from django.utils import timezone
from djautotask.utils import DjautotaskSettings, encode_file_to_base64
from retrying import retry

RETRY_WAIT_EXPONENTIAL_MULTAPPLIER = 1000  # Initial number of milliseconds to
# wait before retrying a request.
RETRY_WAIT_EXPONENTIAL_MAX = 10000  # Maximum number of milliseconds to wait
CACHE_TIMEOUT = 43200
AT_URL_KEY = 'url'
AT_WEB_KEY = 'webUrl'
FORBIDDEN_ERROR_MESSAGE = \
    'The logged in Resource does not have the adequate permissions'
NO_RECORD_ERROR_MESSAGE = \
    'No message. No matching records found. The logged in Resource may not ' \
    'have the required permissions to delete the data.'

logger = logging.getLogger(__name__)


class AutotaskAPIError(Exception):
    """Raise this, not request exceptions."""
    pass


class AutotaskAPIClientError(AutotaskAPIError):
    """
    Raise this to indicate any http error that falls within the
    4xx class of http status codes.
    """
    pass


class AutotaskAPIServerError(AutotaskAPIError):
    """
    Raise this to indicate a Server Error
    500 class of http status codes.
    """
    pass


class AutotaskRecordNotFoundError(AutotaskAPIClientError):
    """The record was not found."""
    pass


class AutotaskSecurityPermissionsException(AutotaskAPIClientError):
    """The API credentials have insufficient security permissions."""
    pass


class AutotaskImpersonationLimitedException(
    AutotaskSecurityPermissionsException
):
    """The impersonation resource has insufficient security permissions."""
    pass


def parse_autotaskapierror(e):
    return ', '.join(e.args)


def retry_if_api_error(exception):
    """
    Return True if we should retry (in this case when it's an
    AutotaskAPIError), False otherwise.

    Basically, don't retry on AutotaskAPIClientError, because those are the
    type of exceptions where retrying won't help (404s, 403s, etc).
    """
    return type(exception) is AutotaskAPIError


def get_cached_url(cache_key):
    return cache.get(f'zone_{cache_key}')


def update_cache(json_obj):
    # set cache timeout as 10 minutes (default: 300)
    cache.set(
        f'zone_{AT_URL_KEY}', json_obj[AT_URL_KEY], timeout=CACHE_TIMEOUT)
    cache.set(
        f'zone_{AT_WEB_KEY}', json_obj[AT_WEB_KEY], timeout=CACHE_TIMEOUT)


def get_api_connection_url(username, force_fetch=False):
    try:
        return _get_connection_url(AT_URL_KEY, username, force_fetch)
    except AutotaskAPIError as e:
        # Save a log even when the get_zone_info request fails.
        SyncJob = apps.get_model('djautotask', 'SyncJob')
        sync_job = SyncJob()
        sync_job.start_time = timezone.now()
        sync_job.message = e
        sync_job.success = False
        sync_job.save()


def get_web_connection_url(username, force_fetch=False):
    return _get_connection_url(AT_WEB_KEY, username, force_fetch)


def _get_connection_url(field, username, force_fetch=False):
    api_url_from_cache = get_cached_url(field)

    if not api_url_from_cache or force_fetch:
        try:
            json_obj = get_zone_info(username)
            # Update cache if empty or forced
            update_cache(json_obj)

            url = json_obj[field]
        except AutotaskAPIError as e:
            raise AutotaskAPIError(f'Failed to get zone info: {e}')
    else:
        url = api_url_from_cache

    return url


def get_zone_info(username):
    endpoint_url = '{}v1.0/zoneInformation?user={}'.format(
        settings.AUTOTASK_SERVER_URL, username
    )

    try:
        logger.debug('Making GET request to {}'.format(endpoint_url))
        response = requests.get(endpoint_url, timeout=3)
        if 200 == response.status_code:
            resp_json = response.json()
            return resp_json
        elif 500 == response.status_code:
            # AT returns 500 if username is blank or incorrect.
            resp_json = response.json()
            raise AutotaskAPIError(
                'Request failed: GET {}: {}'.format(
                    endpoint_url, json.dumps(resp_json)
                )
            )
        else:
            raise AutotaskAPIError(response.content)
    except requests.RequestException as e:
        raise AutotaskAPIError(
            'Request failed: GET {}: {}'.format(
                endpoint_url, e
            )
        )


class ApiCondition:

    def __init__(self, *args, op=None, field=None, value=None):
        self._items = []
        self.op = op
        self.field = field
        self.value = value

        if op in ("and", "or"):
            # Grouping query
            for condition in args:
                if not isinstance(condition, self.__class__):
                    raise TypeError(
                        "Grouped conditions must also be "
                        "instances of {}".format(self.__class__.__name__)
                    )
                self._items.append(condition)

    def __repr__(self):
        if len(self._items):
            return self._items.__repr__()
        return '{{op: {}, field: {}, value: {}}}'.format(
            self.op,
            self.field,
            self.value
        )

    def format_condition(self):
        if len(self._items):
            # Grouping Query
            condition = {
                "op": self.op,
                "items": [
                    api_condition.format_condition()
                    for api_condition in self._items
                ]
            }
        else:
            condition = {
                "op": self.op,
                "field": self.field
            }
            if self.value:
                # value is not required for non-comparison queries
                condition['value'] = self.value

        return condition


class ApiConditionList:
    METHODS = {
        'get': 'query?search=',
        'post': 'query'
    }

    def __init__(self):
        self._list = list()

    def __repr__(self):
        return self._list.__repr__()

    def __len__(self):
        return len(self._list)

    def __getitem__(self, i):
        return self._list[i]

    def __setitem__(self, i, value):
        if not isinstance(value, ApiCondition):
            raise TypeError("Conditions must be instances of ApiCondition.")
        self._list[i] = value

    def __delitem__(self, i):
        del self._list[i]

    def __iter__(self):
        return self._list.__iter__()

    def build_query(self, method="get", **kwargs):
        queries = []
        for condition in self._list:
            queries.append(condition.format_condition())

        condition_filters = {
            'filter': queries,
        }

        max_records = kwargs.get('page_size')
        if max_records:
            condition_filters.update({
                'MaxRecords': max_records
            })

        endpoint = self.METHODS[method]
        filters = json.dumps(condition_filters)

        if method == "get":

            count = kwargs.get('record_count')
            if count:
                endpoint = f'query/count?search={filters}'
            else:
                endpoint += filters

            filters = None
        elif method != "post":
            raise TypeError("Unsupported method")

        return endpoint, filters

    def add(self, condition):
        if not isinstance(condition, ApiCondition):
            raise TypeError("Conditions must be instances of ApiCondition.")

        self._list.append(condition)


class AutotaskAPIClient(object):
    API = None
    MAX_401_ATTEMPTS = 1

    def __init__(
        self,
        username=None,
        password=None,
        integration_code=None,
        rest_api_version=None,
        server_url=None,
        impersonation_resource=None,
    ):
        if not username:
            username = settings.AUTOTASK_CREDENTIALS['username']
        if not password:
            password = settings.AUTOTASK_CREDENTIALS['password']
        if not integration_code:
            integration_code = \
                settings.AUTOTASK_CREDENTIALS['integration_code']
        if not rest_api_version:
            rest_api_version = \
                settings.AUTOTASK_CREDENTIALS['rest_api_version']
        if not server_url:
            server_url = get_api_connection_url(username)

        if not self.API:
            raise ValueError('API not specified')

        self.username = username
        self.password = password
        self.integration_code = integration_code
        self.rest_api_version = rest_api_version
        self.server_url = server_url

        self.request_settings = DjautotaskSettings().get_settings()
        self.timeout = self.request_settings['timeout']
        self.impersonation_resource = impersonation_resource
        self.conditions = ApiConditionList()

        self.cached_body = None

    @property
    def api_base_url(self):
        return '{0}v{1}/'.format(
            self.server_url,
            self.rest_api_version,
        )

    def get_api_url(self, endpoint=None):
        if not endpoint:
            endpoint = self.API

        return '{0}{1}/'.format(
            self.api_base_url,
            endpoint,
        )

    def _log_failed(self, response):
        logger.error('Failed API call: {0} - {1} - {2}'.format(
            response.url, response.status_code, response.content))

    def _prepare_error_for_impersonation(self, msg):
        impersonation_error_msg = \
            'Resource impersonation is enabled in your TopLeft ' \
            'application. Please check the resource security level that is ' \
            'being impersonated.'
        msg = '{} {}'.format(msg, impersonation_error_msg)
        return msg

    def _prepare_error_response(self, response):
        error = response.content.decode("utf-8")
        # decode the bytes encoded error to a string
        # error = error.args[0].decode("utf-8")
        error = error.replace('\r\n', '')

        try:
            error_json = json.loads(error)
            error_list = error_json.get('errors', [])
            if len(error_list) > 1:
                msg = 'Errors: {}'.format(', '.join(error_list))
            elif len(error_list) == 1:
                msg = error_list[0]
            else:
                # No errors given
                msg = 'No error message given.'
        except json.decoder.JSONDecodeError:
            # JSON decoding failed
            msg = 'An error occurred: {} {}'.format(response.status_code,
                                                    error)
        except KeyError:
            # 'code' or 'message' was not found in the error
            msg = 'An error occurred: {} {}'.format(response.status_code,
                                                    error)
        return msg

    def get_headers(self, method):
        headers = {'Content-Type': 'application/json'}

        if self.username:
            headers['UserName'] = self.username
        if self.password:
            headers['Secret'] = self.password
        if self.integration_code:
            headers['ApiIntegrationCode'] = self.integration_code
        if self.impersonation_resource and method.upper() != 'GET':
            # We don't check the impersonation possibility for GET request.
            self.check_impersonation_possibility()
            headers['ImpersonationResourceId'] = str(
                self.impersonation_resource.id
            ) if self.impersonation_resource else None

        return headers

    def _format_fields(self, record, inserted_fields):
        body = {'id': record.id} if record.id else dict()

        for field, value in inserted_fields.items():
            key = field
            body = self._format_request_body(body, key, value)

        return body

    def _format_request_body(self, body, key, value):
        if isinstance(value, datetime.datetime):
            body.update({
                key: value.astimezone(
                    pytz.timezone('UTC')).strftime(
                    "%Y-%m-%dT%H:%M:%SZ")
            })

        elif isinstance(value, models.Model):
            body.update({
                key: value.id
            })

        elif isinstance(value, decimal.Decimal):
            body.update(
                {key: str(value) if value else '0'}
            )

        elif isinstance(value, bool):
            body.update(
                {key: value}
            )
        else:
            body.update(
                {key: str(value) if value else ''}
            )

        return body

    def fetch_resource(self, next_url=None, retry_counter=None, method='get',
                       *args, **kwargs):
        """
        retry_counter is a dict in the form {'count': 0} that is passed in
        to verify the number of attempts that were made.
        """
        @retry(stop_max_attempt_number=self.request_settings['max_attempts'],
               wait_exponential_multiplier=RETRY_WAIT_EXPONENTIAL_MULTAPPLIER,
               wait_exponential_max=RETRY_WAIT_EXPONENTIAL_MAX,
               retry_on_exception=retry_if_api_error)
        def _fetch_resource(endpoint_url, request_retry_counter=None,
                            request_method=method, request_body=None,
                            **kwargs):
            if not request_retry_counter:
                request_retry_counter = {'count': 0}
            request_retry_counter['count'] += 1

            try:
                self.log_message(endpoint_url, request_method, request_body)

                response = requests.request(
                    request_method,
                    endpoint_url,
                    data=request_body,
                    timeout=self.timeout,
                    headers=self.get_headers(request_method),
                )

            except requests.RequestException as e:
                logger.error('Request failed: {} {}: {}'.format(
                    request_method.upper(), endpoint_url, e))
                raise AutotaskAPIError('{}'.format(e))

            if 200 <= response.status_code < 300:
                try:
                    return response.json()
                except JSONDecodeError as e:
                    logger.error(
                        'Request failed during JSON decode: {} {}: {}'.format(
                            request_method.upper(), endpoint_url, e)
                    )
                    raise AutotaskAPIError('{}'.format(e))

            elif response.status_code == 401:
                # It could be the case that zone info has been changed
                msg = 'Unauthorized request: {}'.format(endpoint_url)
                logger.warning(msg)
                if request_retry_counter['count'] <= self.MAX_401_ATTEMPTS:
                    cached_url = get_cached_url(AT_URL_KEY)
                    if cached_url != get_api_connection_url(
                            self.username, force_fetch=True
                    ):
                        logger.info('Zone information has been changed, '
                                    'so this request will be retried.')
                        raise AutotaskAPIError(response.content)
                raise AutotaskAPIClientError(msg)
            elif response.status_code == 403:
                self._log_failed(response)
                raise AutotaskSecurityPermissionsException(
                    self._prepare_error_response(response),
                    response.status_code)
            elif 400 <= response.status_code < 499:
                self._log_failed(response)
                raise AutotaskAPIClientError(
                    self._prepare_error_response(response))
            elif response.status_code == 500:
                self._log_failed(response)
                msg = self._prepare_error_response(response)
                if FORBIDDEN_ERROR_MESSAGE in msg:
                    # Standards, who needs em?
                    raise AutotaskSecurityPermissionsException(msg)
                else:
                    raise AutotaskAPIServerError(
                        self._prepare_error_response(response)
                    )
            else:
                self._log_failed(response)
                raise AutotaskAPIError(
                    self._prepare_error_response(response))

        if next_url:
            url = next_url
        else:
            # Query endpoint is different between GET and POST
            query_endpoint, self.cached_body = \
                self.conditions.build_query(method=method, **kwargs)
            url = "{}{}".format(self.get_api_url(), query_endpoint)

        return _fetch_resource(
            url, request_retry_counter=retry_counter,
            request_method=method, request_body=self.cached_body, **kwargs)

    def log_message(self, endpoint, method, body):
        body = body if body else ''

        logger_message = \
            'Making {} request to {}'.format(method.upper(), endpoint)
        if method == 'post':
            logger_message = \
                '{}. Request body: {}'.format(logger_message, body)

        logger.debug(logger_message)

    def check_impersonation_possibility(self):
        if self.impersonation_resource and not self.impersonation_resource.\
                license_type.has_impersonation(self):
            raise AutotaskImpersonationLimitedException(
                'You do not have the adequate permissions '
                '(impersonation limitation).')

    def request(self, method, endpoint_url, body=None):
        """
        Issue the given type of request to the specified REST endpoint.
        """
        try:
            logger.debug(
                'Making {} request to {}'.format(method, endpoint_url)
            )

            response = requests.request(
                method,
                endpoint_url,
                json=body,
                timeout=self.timeout,
                headers=self.get_headers(method),
            )
        except AutotaskImpersonationLimitedException as e:
            logger.error(
                'Request failed: {} {}: {}'.format(method, endpoint_url, e)
            )
            raise e
        except requests.RequestException as e:
            logger.error(
                'Request failed: {} {}: {}'.format(method, endpoint_url, e)
            )
            raise AutotaskAPIError('{}'.format(e))

        if response.status_code == 204:  # No content
            return None
        elif 200 <= response.status_code < 300:
            return response.json()
        elif response.status_code == 403:
            self._log_failed(response)
            raise AutotaskSecurityPermissionsException(
                self._prepare_error_response(response), response.status_code)
        elif response.status_code == 404:
            msg = 'Resource not found: {}'.format(response.url)
            logger.warning(msg)
            raise AutotaskRecordNotFoundError(msg)
        elif 400 <= response.status_code < 499:
            self._log_failed(response)
            raise AutotaskAPIClientError(
                self._prepare_error_response(response))
        elif response.status_code == 500:
            self._log_failed(response)
            prepared_error = self._prepare_error_response(response)

            # As of Aug. 2021, the REST API returns HTTP 500 for some
            # insufficient permission cases (updating checklist items)
            # even though the docs state that HTTP 403 should be returned.
            # https://webservices2.autotask.net/atservicesrest/swagger/ui/index#/TicketChecklistItemsChild/TicketChecklistItemsChild_CreateEntity # noqa
            # Until this is fixed in the API, check for the message indicating
            # a permission error and raise an appropriate exception.
            # In addition, AT returns 500 when requested record doesn't exist
            # during the deletion
            if FORBIDDEN_ERROR_MESSAGE in prepared_error:
                if self.impersonation_resource:
                    prepared_error = self._prepare_error_for_impersonation(
                        prepared_error
                    )
                raise AutotaskSecurityPermissionsException(prepared_error)
            elif NO_RECORD_ERROR_MESSAGE in prepared_error:
                raise AutotaskRecordNotFoundError(prepared_error, 404)
            else:
                raise AutotaskAPIServerError(prepared_error)
        else:
            self._log_failed(response)
            raise AutotaskAPIError(response)

    def add_condition(self, condition):
        self.conditions.add(condition)
        return len(self.conditions) - 1

    def get_conditions(self):
        return self.conditions

    def remove_condition(self, index):
        del self.conditions[index]

    def clear_conditions(self):
        self.conditions = ApiConditionList()

    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, *args, **kwargs)

    def get_single(self, instance_id):
        endpoint_url = '{}{}'.format(self.get_api_url(), instance_id)
        response = self.fetch_resource(endpoint_url)
        if not response['item']:
            msg = 'Resource not found: {}'.format(endpoint_url)
            logger.warning(msg)
            raise AutotaskRecordNotFoundError(msg)
        return response

    def update(self, instance, changed_fields):
        body = self._format_fields(instance, changed_fields)
        return self.request('patch', self.get_api_url(), body)

    def create(self, instance, **kwargs):
        body = self._format_fields(instance, kwargs)
        # API returns only newly created id
        response = self.request('post', self.get_api_url(), body)
        return response.get('itemId')

    def delete(self, instance, parent=None):
        endpoint_url = self.get_api_url() + str(instance.id)
        response = self.request('delete', endpoint_url)
        # AT sends deleted_id or 500 in the case failure instead of 404 or 204
        return response.get('itemId')


class ChildAPIMixin:
    PARENT_API = None
    CHILD_API = None

    def get_child_url(self, parent_id):
        return '{}{}/{}/{}'.format(
            self.api_base_url,
            self.PARENT_API,
            parent_id,
            self.CHILD_API
        )

    def update(self, instance, parent, changed_fields):
        endpoint_url = self.get_child_url(parent.id)
        body = self._format_fields(instance, changed_fields)
        return self.request('patch', endpoint_url, body)

    def create(self, instance, parent, **kwargs):
        endpoint_url = self.get_child_url(parent.id)
        body = self._format_fields(instance, kwargs)
        response = self.request('post', endpoint_url, body)
        return response.get('itemId')

    def delete(self, instance, parent):
        endpoint_url = '{}/{}'.format(
            self.get_child_url(parent.id),
            str(instance.id)
        )
        response = self.request('delete', endpoint_url)
        # AT sends deleted_id or 500 in the case failure instead of 404 or 204
        return response.get('itemId')

    # use POST method because of IN-clause query string
    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, method='post', *args, **kwargs)


class ContactsAPIClient(AutotaskAPIClient):
    API = 'Contacts'


class RolesAPIClient(AutotaskAPIClient):
    API = 'Roles'


class DepartmentsAPIClient(AutotaskAPIClient):
    API = 'Departments'


class ResourceServiceDeskRolesAPIClient(AutotaskAPIClient):
    API = 'ResourceServiceDeskRoles'


class ResourceRoleDepartmentsAPIClient(AutotaskAPIClient):
    API = 'ResourceRoleDepartments'


class TicketsAPIClient(AutotaskAPIClient):
    API = 'Tickets'

    def count(self, next_url, *args, **kwargs):
        # Make get request using Api conditions
        return self.fetch_resource(next_url, *args, **kwargs)


class BillingCodesAPIClient(AutotaskAPIClient):
    API = 'BillingCodes'


class ContractsAPIClient(AutotaskAPIClient):
    API = 'Contracts'


class ContractExclusionSetAPIClient(AutotaskAPIClient):
    API = 'ContractExclusionSets'


class ContractsExcludedWorkTypeAPIClient(AutotaskAPIClient):
    API = 'ContractExclusionSetExcludedWorkTypes'


class ContractsExcludedRoleAPIClient(AutotaskAPIClient):
    API = 'ContractExclusionSetExcludedRoles'


class AccountPhysicalLocationsAPIClient(AutotaskAPIClient):
    API = 'CompanyLocations'

    # use POST method because of IN-clause query string
    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, method='post', *args, **kwargs)


class TasksAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'Tasks'
    PARENT_API = 'Projects'
    CHILD_API = API
    # For tasks, child API endpoint is the same


class TicketNotesAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'TicketNotes'
    PARENT_API = 'Tickets'
    CHILD_API = 'Notes'


class TaskNotesAPIClient(AutotaskAPIClient):
    API = 'TaskNotes'

    # use POST method because of IN-clause query string
    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, method='post', *args, **kwargs)


class TimeEntriesAPIClient(AutotaskAPIClient):
    API = 'TimeEntries'

    # use POST method because of IN-clause query string
    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, method='post', *args, **kwargs)


class TicketSecondaryResourcesAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'TicketSecondaryResources'
    PARENT_API = 'Tickets'
    CHILD_API = 'SecondaryResources'


class TaskSecondaryResourcesAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'TaskSecondaryResources'
    PARENT_API = 'Tasks'
    CHILD_API = 'SecondaryResources'


class ProjectsAPIClient(AutotaskAPIClient):
    API = 'Projects'

    # use POST method because of IN-clause query string
    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, method='post', *args, **kwargs)


class TicketCategoriesAPIClient(AutotaskAPIClient):
    API = 'TicketCategories'


class CompanyAlertAPIClient(AutotaskAPIClient):
    API = 'CompanyAlerts'


class ProjectNotesAPIClient(AutotaskAPIClient):
    API = 'ProjectNotes'


class TaskPredecessorsAPIClient(AutotaskAPIClient):
    API = 'TaskPredecessors'

    # use POST method because of IN-clause query string
    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, method='post', *args, **kwargs)


class ServiceCallsAPIClient(AutotaskAPIClient):
    API = 'ServiceCalls'

    # use POST method because of IN-clause query string
    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, method='post', *args, **kwargs)


class ServiceCallTicketsAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'ServiceCallTickets'
    PARENT_API = 'ServiceCalls'
    CHILD_API = 'Tickets'


class ServiceCallTicketResourcesAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'ServiceCallTicketResources'
    PARENT_API = 'ServiceCallTickets'
    CHILD_API = 'Resources'


class ServiceCallTasksAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'ServiceCallTasks'
    PARENT_API = 'ServiceCalls'
    CHILD_API = 'Tasks'


class ServiceCallTaskResourcesAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'ServiceCallTaskResources'
    PARENT_API = 'ServiceCallTasks'
    CHILD_API = 'Resources'


class ResourcesAPIClient(AutotaskAPIClient):
    API = 'Resources'


class AccountsAPIClient(AutotaskAPIClient):
    API = 'Companies'


class PhasesAPIClient(AutotaskAPIClient):
    API = 'Phases'


class UDFAPIClient(AutotaskAPIClient):

    def get_api_url(self, endpoint=None):
        return '{0}{1}/entityInformation/userDefinedFields'.format(
            self.api_base_url,
            self.API,
        )


class TicketsUDFAPIClient(UDFAPIClient):
    API = 'Tickets'


class TasksUDFAPIClient(UDFAPIClient):
    API = 'Tasks'


class ProjectsUDFAPIClient(UDFAPIClient):
    API = 'Projects'


class AutotaskPicklistAPIClient(AutotaskAPIClient):

    def __init__(self, **kwargs):
        self.API = '{}/entityInformation/fields'.format(self.API_ENTITY)
        super().__init__(**kwargs)

    def get(self, next_url, *args, **kwargs):
        # Get either the next_url provided by the AT response or build the
        # initial url.
        return self.fetch_resource(
            next_url or self.get_api_url(), *args, **kwargs)


class NoteTypesAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'TicketNotes'


class LicenseTypesAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'Resources'


class UseTypesAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'BillingCodes'


class BillingCodeTypeAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'BillingCodes'


class TaskTypeLinksAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'TimeEntries'


class AccountTypesAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'Companies'


class TicketCategoryPicklistAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'TicketCategories'


class ServiceCallStatusPicklistAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'ServiceCalls'


class TicketPicklistAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'Tickets'


class TaskPicklistAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'Tasks'


class ProjectPicklistAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'Projects'


class ProjectNoteTypesAPIClient(AutotaskPicklistAPIClient):
    API_ENTITY = 'ProjectNotes'


class TicketChecklistItemsAPIClient(ChildAPIMixin, AutotaskAPIClient):
    API = 'TicketChecklistItems'
    PARENT_API = 'Tickets'
    CHILD_API = 'ChecklistItems'

    def get(self, next_url, *args, **kwargs):
        return self.fetch_resource(next_url, *args, **kwargs)

    def update(self, parent, **kwargs):
        endpoint_url = self.get_child_url(parent)
        return self.request('patch', endpoint_url, kwargs)

    def create(self, parent, **kwargs):
        endpoint_url = self.get_child_url(parent)
        return self.request('post', endpoint_url, kwargs)

    def delete(self, parent, **kwargs):
        endpoint_url = '{}/{}'.format(
            self.get_child_url(parent),
            kwargs.get("id")
        )
        return self.request('delete', endpoint_url)


class AttachmentInfoAPIClient(AutotaskAPIClient):
    API = 'AttachmentInfo'

    def document_download(self, object_id, document_id, record_type):
        ENDPOINT_DOCUMENTS_DOWNLOAD = f'{record_type}/{object_id}' \
                                      f'/Attachments/{document_id}'

        endpoint = f'{self.api_base_url}/{ENDPOINT_DOCUMENTS_DOWNLOAD}'

        try:
            logger.debug('Making GET request to {}'.format(endpoint))
            response = requests.get(
                endpoint,
                timeout=self.timeout,
                headers=self.get_headers('GET'),
            )
        except requests.RequestException as e:
            logger.error('Request failed: GET {}: {}'.format(endpoint, e))
            raise AutotaskAPIError('{}'.format(e))

        if 200 <= response.status_code < 300:
            if response.json().get('items'):
                response = response.json().get('items')[0]

            return response
        else:
            self._log_failed(response)
            return None

    def upload_attachments(self, object_id, file, recordType):
        ENDPOINT_DOCUMENTS = f'{recordType}/{object_id}/Attachments'
        endpoint_url = f'{self.api_base_url}/{ENDPOINT_DOCUMENTS}'

        file_name, file_content = file['file']
        file_data = encode_file_to_base64(file_content)

        body = {
            "id": 0,
            "attachmentType": "FILE_ATTACHMENT",
            "fullPath": file_name,
            "publish": 1,
            "title": file_name,
            "data": file_data
        }

        response = self.request('POST', endpoint_url, body)

        return response

    def get_attachment(self, object_id, document_id, record_type):
        return self.document_download(object_id, document_id, record_type)

    def count(self, *args, **kwargs):
        kwargs['record_count'] = True

        return self.fetch_resource(*args, **kwargs)



# Generated by Django 4.2.16 on 2024-10-08 02:10

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0117_alter_tasknote_task_alter_ticketnote_ticket'),
    ]

    operations = [
        migrations.AlterModelTable(
            name='tasktypetracker',
            table=None,
        ),
        migrations.CreateModel(
            name='CompanyAlert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('alert_text', models.TextField(blank=True, max_length=8000, null=True)),
                ('alert_type', models.IntegerField(blank=True, null=True)),
                ('account', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='alerts', to='djautotask.account')),
            ],
        ),
        migrations.CreateModel(
            name='CompanyAlertTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_companyalert',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.companyalert',),
        ),
    ]



# Generated by Django 2.1.14 on 2020-01-31 16:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0039_merge_20200131_1022'),
        ('djautotask', '0039_auto_20200131_1134'),
    ]

    operations = [
    ]



# Generated by Django 2.1.11 on 2019-09-23 09:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0009_auto_20190918_1525'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='project',
            name='completed_date_time',
        ),
        migrations.RemoveField(
            model_name='project',
            name='end_date_time',
        ),
        migrations.RemoveField(
            model_name='project',
            name='start_date_time',
        ),
        migrations.AddField(
            model_name='project',
            name='completed_date',
            field=models.DateField(null=True),
        ),
        migrations.AddField(
            model_name='project',
            name='end_date',
            field=models.DateField(null=True),
        ),
        migrations.AddField(
            model_name='project',
            name='start_date',
            field=models.DateField(null=True),
        ),
    ]



# Generated by Django 4.2.16 on 2024-10-24 15:21

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0118_alter_tasktypetracker_table_companyalert_and_more'),
        ('djautotask', '0118_alter_ticket_due_date_time'),
    ]

    operations = [
    ]



# Generated by Django 4.2.16 on 2024-10-23 14:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0117_alter_tasknote_task_alter_ticketnote_ticket'),
    ]

    operations = [
        migrations.AlterField(
            model_name='ticket',
            name='due_date_time',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



# Generated by Django 4.0.7 on 2023-03-01 13:11

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0111_alter_phase_estimated_hours'),
    ]

    operations = [
        migrations.AddField(
            model_name='account',
            name='owner_resource',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djautotask.resource'),
        ),
    ]



# Generated by Django 2.1.14 on 2019-11-29 15:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0031_auto_20191129_1454'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='displaycolor',
            options={'ordering': ('label',), 'verbose_name_plural': 'Display colors'},
        ),
        migrations.AlterModelOptions(
            name='projectstatus',
            options={'ordering': ('label',), 'verbose_name_plural': 'Project statuses'},
        ),
        migrations.AlterModelOptions(
            name='queue',
            options={'ordering': ('label',), 'verbose_name_plural': 'Queues'},
        ),
        migrations.AlterModelOptions(
            name='status',
            options={'ordering': ('label',), 'verbose_name_plural': 'Statuses'},
        ),
    ]



# Generated by Django 3.1.2 on 2021-01-08 12:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0086_auto_20201222_1402'),
    ]

    operations = [
        migrations.AlterField(
            model_name='accounttype',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='displaycolor',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='issuetype',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='licensetype',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='notetype',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='priority',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='projectstatus',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='projecttype',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='queue',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='servicecallstatus',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='source',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='status',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='subissuetype',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='tasktypelink',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='tickettype',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='usetype',
            name='sort_order',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.1.11 on 2020-04-16 12:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0060_auto_20200414_1522'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='servicecallstatus',
            options={'ordering': ('label',), 'verbose_name_plural': 'Service call statuses'},
        ),
    ]



# Generated by Django 4.0 on 2022-02-17 15:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0103_auto_20220106_1622'),
    ]

    operations = [
        migrations.DeleteModel(
            name='AllocationCodeTracker',
        ),
        migrations.RenameModel(
            old_name='AllocationCode',
            new_name='BillingCode',
        ),
        migrations.CreateModel(
            name='BillingCodeTracker',
            fields=[
            ],
            options={
                'db_table': 'djautotask_billingcode',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('djautotask.billingcode',),
        ),
        migrations.RenameField(
            model_name='task',
            old_name='allocation_code',
            new_name='billing_code',
        ),
        migrations.RenameField(
            model_name='ticket',
            old_name='allocation_code',
            new_name='billing_code',
        ),
        migrations.RenameField(
            model_name='timeentry',
            old_name='allocation_code',
            new_name='billing_code',
        ),
    ]



# Generated by Django 2.1.14 on 2020-07-14 08:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0069_auto_20200602_1419'),
    ]

    operations = [
        migrations.AlterField(
            model_name='queue',
            name='label',
            field=models.CharField(blank=True, max_length=100, null=True),
        ),
    ]



# Generated by Django 4.2.11 on 2024-06-17 10:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0115_task_task_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='servicecall',
            name='tasks',
            field=models.ManyToManyField(through='djautotask.ServiceCallTask', to='djautotask.task'),
        ),
        migrations.AlterField(
            model_name='servicecall',
            name='tickets',
            field=models.ManyToManyField(through='djautotask.ServiceCallTicket', to='djautotask.ticket'),
        ),
    ]
